from __future__ import annotations

import copy
import gc
import logging
import multiprocessing as mp
from pathlib import Path
from typing import Callable, Literal

import numpy as np
import scipy.optimize
from lgdo import Array, Histogram, Scalar, lh5
from numba import njit
from numpy.typing import NDArray

from ..log_utils import setup_log
from .evt import EVT_TABLE_NAME, read_optmap_evt
from .optmap import OpticalMap

log = logging.getLogger(__name__)


def _optmaps_for_channels(
    optmap_evt_columns: list[str],
    settings,
    chfilter: tuple[str | int] | Literal["*"] = (),
    use_shmem: bool = False,
):
    all_det_ids = [ch_id for ch_id in optmap_evt_columns if ch_id.isnumeric()]

    if chfilter != "*":
        chfilter = [str(ch) for ch in chfilter]  # normalize types
        optmap_det_ids = [det for det in all_det_ids if str(det) in chfilter]
    else:
        optmap_det_ids = all_det_ids

    log.info("creating empty optmaps")
    optmap_count = len(optmap_det_ids) + 1
    optmaps = [
        OpticalMap("all" if i == 0 else optmap_det_ids[i - 1], settings, use_shmem)
        for i in range(optmap_count)
    ]

    return all_det_ids, optmaps, optmap_det_ids


@njit(cache=True)
def _compute_hit_maps(hitcounts, optmap_count, ch_idx_to_optmap):
    mask = np.zeros((hitcounts.shape[0], optmap_count), dtype=np.bool_)
    counts = hitcounts.sum(axis=1)
    for idx in range(hitcounts.shape[0]):
        if counts[idx] == 0:
            continue

        for ch_idx in range(hitcounts.shape[1]):
            c = hitcounts[idx, ch_idx]
            if c > 0:  # detected
                mask[idx, 0] = True
                mask_idx = ch_idx_to_optmap[ch_idx]
                if mask_idx > 0:
                    mask[idx, mask_idx] = True
    return mask


def _fill_hit_maps(optmaps: list[OpticalMap], loc, hitcounts: NDArray, ch_idx_to_map_idx):
    masks = _compute_hit_maps(hitcounts, len(optmaps), ch_idx_to_map_idx)

    for i in range(len(optmaps)):
        locm = loc[masks[:, i]]
        optmaps[i].fill_hits(locm)


def _count_multi_ph_detection(hitcounts) -> NDArray:
    hits_per_primary = hitcounts.sum(axis=1)
    bins = np.arange(0, hits_per_primary.max() + 1.5) - 0.5
    return np.histogram(hits_per_primary, bins)[0]


def _fit_multi_ph_detection(hits_per_primary) -> float:
    if len(hits_per_primary) <= 2:  # have only 0 and 1 hits, can't fit (and also don't need to).
        return np.inf

    x = np.arange(0, len(hits_per_primary))
    popt, pcov = scipy.optimize.curve_fit(
        lambda x, p0, k: p0 * np.exp(-k * x), x[1:], hits_per_primary[1:]
    )
    best_fit_exponent = popt[1]

    log.info(
        "p(> 1 detected photon)/p(1 detected photon) = %f",
        sum(hits_per_primary[2:]) / hits_per_primary[1],
    )
    log.info(
        "p(> 1 detected photon)/p(<=1 detected photon) = %f",
        sum(hits_per_primary[2:]) / sum(hits_per_primary[0:2]),
    )

    return best_fit_exponent


def _create_optical_maps_process_init(optmaps, log_level) -> None:
    # need to use shared global state. passing the shared memory arrays via "normal" arguments to
    # the worker function is not supported...
    global _shared_optmaps  # noqa: PLW0603
    _shared_optmaps = optmaps

    # setup logging in the worker process.
    setup_log(log_level, multiproc=True)


def _create_optical_maps_process(
    optmap_events_fn, buffer_len, all_det_ids, ch_idx_to_map_idx
) -> None:
    log.info("started worker task for %s", optmap_events_fn)
    x = _create_optical_maps_chunk(
        optmap_events_fn,
        buffer_len,
        all_det_ids,
        _shared_optmaps,
        ch_idx_to_map_idx,
    )
    log.info("finished worker task for %s", optmap_events_fn)
    return tuple(int(i) for i in x)


def _create_optical_maps_chunk(
    optmap_events_fn, buffer_len, all_det_ids, optmaps, ch_idx_to_map_idx
) -> None:
    optmap_events_it = read_optmap_evt(optmap_events_fn, buffer_len)

    hits_per_primary = np.zeros(10, dtype=np.int64)
    hits_per_primary_len = 0
    for it_count, (events_lgdo, events_entry, event_n_rows) in enumerate(optmap_events_it):
        assert (it_count == 0) == (events_entry == 0)
        optmap_events = events_lgdo.view_as("pd").iloc[0:event_n_rows]
        hitcounts = optmap_events[all_det_ids].to_numpy()
        loc = optmap_events[["xloc", "yloc", "zloc"]].to_numpy()

        log.debug("filling vertex histogram (%d)", it_count)
        optmaps[0].fill_vertex(loc)

        log.debug("filling hits histogram (%d)", it_count)
        _fill_hit_maps(optmaps, loc, hitcounts, ch_idx_to_map_idx)
        hpp = _count_multi_ph_detection(hitcounts)
        hits_per_primary_len = max(hits_per_primary_len, len(hpp))
        hits_per_primary[0 : len(hpp)] += hpp

    # commit the final part of the hits to the maps.
    for i in range(len(optmaps)):
        optmaps[i].fill_hits_flush()
        gc.collect()

    return hits_per_primary[0:hits_per_primary_len]


def create_optical_maps(
    optmap_events_fn: list[str],
    settings,
    buffer_len: int = int(5e6),
    chfilter: tuple[str | int] | Literal["*"] = (),
    output_lh5_fn: str | None = None,
    after_save: Callable[[int, str, OpticalMap]] | None = None,
    check_after_create: bool = False,
    n_procs: int | None = 1,
) -> None:
    """Create optical maps.

    Parameters
    ----------
    optmap_events_fn
        list of filenames to lh5 files with a table ``/optmap_evt`` with columns ``{x,y,z}loc``
        and one column (with numeric header) for each SiPM channel.
    chfilter
        tuple of detector ids that will be included in the resulting optmap. Those have to match
        the column names in ``optmap_events_fn``.
    n_procs
        number of processors, ``1`` for sequential mode, or ``None`` to use all processors.
    """
    if len(optmap_events_fn) == 0:
        msg = "no input files specified"
        raise ValueError(msg)

    use_shmem = n_procs is None or n_procs > 1

    optmap_evt_columns = list(
        lh5.read(EVT_TABLE_NAME, optmap_events_fn[0], start_row=0, n_rows=1).keys()
    )  # peek into the (first) file to find column names.
    all_det_ids, optmaps, optmap_det_ids = _optmaps_for_channels(
        optmap_evt_columns, settings, chfilter=chfilter, use_shmem=use_shmem
    )

    # indices for later use in _compute_hit_maps.
    ch_idx_to_map_idx = np.array(
        [optmap_det_ids.index(d) + 1 if d in optmap_det_ids else -1 for d in all_det_ids]
    )
    assert np.sum(ch_idx_to_map_idx > 0) == len(optmaps) - 1

    log.info("creating optical map groups: %s", ", ".join(["all", *optmap_det_ids]))

    q = []

    # sequential mode.
    if not use_shmem:
        for fn in optmap_events_fn:
            q.append(
                _create_optical_maps_chunk(fn, buffer_len, all_det_ids, optmaps, ch_idx_to_map_idx)
            )
    else:
        ctx = mp.get_context("forkserver")
        for i in range(len(optmaps)):
            optmaps[i]._mp_preinit(ctx, vertex=(i == 0))

        # note: errors thrown in initializer will make the main process hang in an endless loop.
        # unfortunately, we cannot pass the objects later, as they contain shmem/array handles.
        pool = ctx.Pool(
            n_procs,
            initializer=_create_optical_maps_process_init,
            initargs=(optmaps, log.getEffectiveLevel()),
            maxtasksperchild=1,  # re-create worker after each task, to avoid leaking memory.
        )

        pool_results = []
        for fn in optmap_events_fn:
            r = pool.apply_async(
                _create_optical_maps_process,
                args=(fn, buffer_len, all_det_ids, ch_idx_to_map_idx),
            )
            pool_results.append((r, fn))

        pool.close()
        for r, fn in pool_results:
            try:
                q.append(np.array(r.get()))
            except BaseException as e:
                msg = f"error while processing file {fn}"
                raise RuntimeError(msg) from e  # re-throw errors of workers.
        log.debug("got all worker results")
        pool.join()
        log.info("joined worker process pool")

    # merge hitcounts.
    if len(q) != len(optmap_events_fn):
        log.error("got %d results for %d files", len(q), len(optmap_events_fn))
    hits_per_primary = np.zeros(10, dtype=np.int64)
    hits_per_primary_len = 0
    for hitcounts in q:
        hits_per_primary[0 : len(hitcounts)] += hitcounts
        hits_per_primary_len = max(hits_per_primary_len, len(hitcounts))

    hits_per_primary = hits_per_primary[0:hits_per_primary_len]
    hits_per_primary_exponent = _fit_multi_ph_detection(hits_per_primary)

    # all maps share the same vertex histogram.
    for i in range(1, len(optmaps)):
        optmaps[i].h_vertex = optmaps[0].h_vertex

    log.info("computing probability and storing to %s", output_lh5_fn)
    for i in range(len(optmaps)):
        optmaps[i].create_probability()
        if check_after_create:
            optmaps[i].check_histograms()
        group = "all" if i == 0 else "_" + optmap_det_ids[i - 1]
        if output_lh5_fn is not None:
            optmaps[i].write_lh5(lh5_file=output_lh5_fn, group=group)

        if after_save is not None:
            after_save(i, group, optmaps[i])

        optmaps[i] = None  # clear some memory.

    if output_lh5_fn is not None:
        lh5.write(Array(hits_per_primary), "_hitcounts", lh5_file=output_lh5_fn)
        lh5.write(Scalar(hits_per_primary_exponent), "_hitcounts_exp", lh5_file=output_lh5_fn)


def list_optical_maps(lh5_file: str) -> list[str]:
    maps = lh5.ls(lh5_file)
    return [m for m in maps if m not in ("_hitcounts", "_hitcounts_exp")]


def _merge_optical_maps_process(
    d: str,
    map_l5_files: list[str],
    output_lh5_fn: str,
    settings,
    check_after_create: bool = False,
    write_part_file: bool = False,
) -> bool:
    log.info("merging optical map group: %s", d)
    merged_map = OpticalMap.create_empty(d, settings)
    merged_nr_gen = merged_map.h_vertex
    merged_nr_det = merged_map.h_hits

    all_edges = None
    for optmap_fn in map_l5_files:
        nr_det = lh5.read(f"/{d}/nr_det", optmap_fn)
        assert isinstance(nr_det, Histogram)
        nr_gen = lh5.read(f"/{d}/nr_gen", optmap_fn)
        assert isinstance(nr_gen, Histogram)

        assert OpticalMap._edges_eq(nr_det.binning, nr_gen.binning)
        if all_edges is not None and not OpticalMap._edges_eq(nr_det.binning, all_edges):
            msg = "edges of input optical maps differ"
            raise ValueError(msg)
        all_edges = nr_det.binning

        # now that we validated that the map dimensions are equal, add up the actual data (in counts).
        merged_nr_det += nr_det.weights.nda
        merged_nr_gen += nr_gen.weights.nda

    merged_map.create_probability()
    if check_after_create:
        merged_map.check_histograms(include_prefix=True)

    if write_part_file:
        output_lh5_fn = f"{output_lh5_fn}_{d}.mappart.lh5"
    wo_mode = "overwrite_file" if write_part_file else "write_safe"
    merged_map.write_lh5(lh5_file=output_lh5_fn, group=d, wo_mode=wo_mode)

    return output_lh5_fn


def merge_optical_maps(
    map_l5_files: list[str],
    output_lh5_fn: str,
    settings,
    check_after_create: bool = False,
    n_procs: int | None = 1,
) -> None:
    """Merge optical maps from multiple files.

    Parameters
    ----------
    n_procs
        number of processors, ``1`` for sequential mode, or ``None`` to use all processors.
    """
    # verify that we have the same maps in all files.
    all_det_ntuples = None
    for optmap_fn in map_l5_files:
        det_ntuples = list_optical_maps(optmap_fn)
        if all_det_ntuples is not None and det_ntuples != all_det_ntuples:
            msg = "available optical maps in input files differ"
            raise ValueError(msg)
        all_det_ntuples = det_ntuples

    log.info("merging optical map groups: %s", ", ".join(all_det_ntuples))

    use_mp = (n_procs is None or n_procs > 1) and len(all_det_ntuples) > 1

    if not use_mp:
        # sequential mode: merge maps one-by-one.
        for d in all_det_ntuples:
            _merge_optical_maps_process(
                d, map_l5_files, output_lh5_fn, settings, check_after_create, use_mp
            )
    else:
        ctx = mp.get_context("forkserver")

        # note: errors thrown in initializer will make the main process hang in an endless loop.
        pool = ctx.Pool(
            n_procs,
            initializer=_create_optical_maps_process_init,
            initargs=(None, log.getEffectiveLevel()),
            maxtasksperchild=1,  # re-create worker after each task, to avoid leaking memory.
        )

        pool_results = []

        # merge maps in workers.
        for d in all_det_ntuples:
            r = pool.apply_async(
                _merge_optical_maps_process,
                args=(d, map_l5_files, output_lh5_fn, settings, check_after_create, use_mp),
            )
            pool_results.append((r, d))

        pool.close()
        q = []
        for r, d in pool_results:
            try:
                q.append((d, r.get()))
            except BaseException as e:
                msg = f"error while processing map {d}"
                raise RuntimeError(msg) from e  # re-throw errors of workers.

        log.debug("got all worker results")
        pool.join()
        log.info("joined worker process pool")

        # transfer to actual output file.
        for d, part_fn in q:
            assert isinstance(part_fn, str)
            for h_name in ("nr_det", "nr_gen", "p_det", "p_det_err"):
                obj = f"/{d}/{h_name}"
                log.info("transfer %s from %s", obj, part_fn)
                h = lh5.read(obj, part_fn)
                assert isinstance(h, Histogram)
                lh5.write(h, obj, output_lh5_fn, wo_mode="write_safe")
            Path(part_fn).unlink()

    # merge hitcounts.
    hits_per_primary = np.zeros(10, dtype=np.int64)
    hits_per_primary_len = 0
    for optmap_fn in map_l5_files:
        hitcounts = lh5.read("/_hitcounts", optmap_fn)
        assert isinstance(hitcounts, Array)
        hits_per_primary[0 : len(hitcounts)] += hitcounts
        hits_per_primary_len = max(hits_per_primary_len, len(hitcounts))

    hits_per_primary = hits_per_primary[0:hits_per_primary_len]
    lh5.write(Array(hits_per_primary), "_hitcounts", lh5_file=output_lh5_fn)

    # re-calculate hitcounts exponent.
    hits_per_primary_exponent = _fit_multi_ph_detection(hits_per_primary)
    lh5.write(Scalar(hits_per_primary_exponent), "_hitcounts_exp", lh5_file=output_lh5_fn)


def check_optical_map(map_l5_file: str):
    """Run a health check on the map file.

    This checks for consistency, and output details on map statistics.
    """
    if lh5.read("_hitcounts_exp", lh5_file=map_l5_file).value != np.inf:
        log.error("unexpected hitcount exp not equal to positive infinity")
        return
    if lh5.read("_hitcounts", lh5_file=map_l5_file).nda.shape != (2,):
        log.error("unexpected hitcount shape")
        return

    all_binning = None
    for submap in list_optical_maps(map_l5_file):
        try:
            om = OpticalMap.load_from_file(map_l5_file, submap)
        except Exception:
            log.exception("error while loading optical map %s", submap)
            continue
        om.check_histograms(include_prefix=True)

        if all_binning is not None and not OpticalMap._edges_eq(om.binning, all_binning):
            log.error("edges of optical map %s differ", submap)
        else:
            all_binning = om.binning


def rebin_optical_maps(map_l5_file: str, output_lh5_file: str, factor: int):
    """Rebin the optical map by an integral factor.

    .. note ::

        the factor has to divide the bincounts on all axes.
    """
    if not isinstance(factor, int) or factor <= 1:
        msg = f"invalid rebin factor {factor}"
        raise ValueError(msg)

    def _rebin_map(large: NDArray, factor: int) -> NDArray:
        factor = np.full(3, factor, dtype=int)
        sh = np.column_stack([np.array(large.shape) // factor, factor]).ravel()
        return large.reshape(sh).sum(axis=(1, 3, 5))

    for submap in list_optical_maps(map_l5_file):
        log.info("rebinning optical map group: %s", submap)

        om = OpticalMap.load_from_file(map_l5_file, submap)

        settings = om.get_settings()
        if not all(b % factor == 0 for b in settings["bins"]):
            msg = f"invalid factor {factor}, not a divisor"
            raise ValueError(msg)
        settings = copy.copy(settings)
        settings["bins"] = [b // factor for b in settings["bins"]]

        om_new = OpticalMap.create_empty(om.name, settings)
        om_new.h_vertex = _rebin_map(om.h_vertex, factor)
        om_new.h_hits = _rebin_map(om.h_hits, factor)
        om_new.create_probability()
        om_new.write_lh5(lh5_file=output_lh5_file, group=submap, wo_mode="write_safe")

    # just copy hitcounts exponent.
    for dset in ("_hitcounts_exp", "_hitcounts"):
        lh5.write(lh5.read(dset, lh5_file=map_l5_file), dset, lh5_file=output_lh5_file)



from __future__ import annotations

import importlib
import itertools
import logging
import re
from collections.abc import Iterable, Mapping
from contextlib import contextmanager
from pathlib import Path

from dbetto import AttrsDict
from lgdo.types import Table, VectorOfVectors

log = logging.getLogger(__name__)


def get_wo_mode(
    group: int, out_det: int, in_det: int, chunk: int, new_hit_file: bool, overwrite: bool = False
):
    """Get the mode for lh5 file writing."""
    indices = [group, out_det, in_det, chunk]

    good_idx = all(i == 0 for i in indices)

    if good_idx and new_hit_file:
        return "of" if overwrite else "w"

    # if we have a detector not the first and chunk 0 append column
    if ((in_det > 0) or (out_det > 0)) & (chunk == 0):
        return "ac"
    return "a"


def get_file_dict(
    stp_files: list[str] | str,
    glm_files: list[str] | str | None,
    hit_files: list[str] | str | None = None,
) -> AttrsDict:
    """Get the file info as a AttrsDict.

    Creates an :class:`dbetto.AttrsDict` with keys `stp_files`,
    `glm_files` and `hit_files`. Each key contains a list of
    file-paths (or `None`).

    Parameters
    ----------
    stp_files
        string or list of strings of the stp files.
    glm_files
        string or list of strings of the glm files, or None in which
        case the glm will be created in memory.
    hit_files
        string or list of strings of the hit files, if None the output
        files will be created in memory.
    """
    # make a list of the right length
    if isinstance(stp_files, str):
        stp_files = [stp_files]

    glm_files_list = [None] * len(stp_files) if glm_files is None else glm_files

    # make a list of files in case
    # 1) hit_files is a str and stp_files is a list
    # 2) hit_files and stp_files are both lists of different length

    hit_is_list = isinstance(hit_files, list)

    if not hit_is_list:
        hit_files_list = [hit_files] * len(stp_files)
    elif hit_is_list and len(hit_files) == 1 and len(stp_files) > 1:
        hit_files_list = [hit_files[0]] * len(stp_files)
    else:
        hit_files_list = hit_files

    files = {}

    for file_type, file_list in zip(
        ["stp", "glm", "hit"], [stp_files, glm_files_list, hit_files_list]
    ):
        if isinstance(file_list, str):
            files[file_type] = [file_list]
        else:
            files[file_type] = file_list

    return AttrsDict(files)


def get_file_list(path: str | None, threads: int | None = None) -> list[str]:
    """Get a list of files accounting for the multithread index."""
    if threads is None or path is None:
        return path
    return [f"{(Path(path).with_suffix(''))}_t{idx}.lh5" for idx in range(threads)]


def copy_units(tab: Table) -> dict:
    """Extract a dictionary of attributes (i.e. units).

    Parameters
    ----------
    tab
        Table to get the units from.

    Returns
    -------
    a dictionary with the units for each field
    in the table.
    """
    units = {}

    for field in tab:
        if "units" in tab[field].attrs:
            units[field] = tab[field].attrs["units"]

    return units


def assign_units(tab: Table, units: Mapping) -> Table:
    """Copy the attributes from the map of attributes to the table.

    Parameters
    ----------
    tab
        Table to add attributes to.
    units
        mapping (dictionary like) of units of each field

    Returns
    -------
    an updated table with LGDO attributes.
    """
    for field in tab:
        if field in units:
            if not isinstance(tab[field], VectorOfVectors):
                tab[field].attrs["units"] = units[field]
            else:
                tab[field].flattened_data.attrs["units"] = units[field]

    return tab


def _search_string(string: str):
    """Capture the characters matching the pattern for a function call."""
    pattern = r"\b([a-zA-Z_][a-zA-Z0-9_\.]*)\s*\("
    return re.findall(pattern, string)


def get_function_string(expr: str, aliases: dict | None = None) -> tuple[str, dict]:
    """Get a function call to evaluate.

    Search for any patterns matching the pattern for a function call.
    We also detect any cases of aliases being used, by default
    just for `numpy` as `np` and `awkward` as `ak`. In this
    case, the full name is replaces with the alias in the expression
    and also in the output globals dictionary.

    It is possible to chain together functions eg:

    .. code-block:: python

        ak.num(np.array([1, 2]))

    and all packages will be imported.

    Parameters
    ----------
    expr
        expression to evaluate.
    aliases
        dictionary of package aliases for names used in dictionary. These allow to
        give shorter names to packages. This is combined with two defaults `ak` for
        `awkward` and `np` for `numpy`. If `None` is supplied only these are used.

    Returns
    -------
    a tuple of call string and dictionary of the imported global packages.
    """
    # aliases for easier lookup
    aliases = (
        {"numpy": "np", "awkward": "ak"}
        if aliases is None
        else aliases | {"numpy": "np", "awkward": "ak"}
    )

    # move to only alias names
    for name, short_name in aliases.items():
        expr = expr.replace(name, short_name)

    globs = {}
    # search on the whole expression

    funcs = _search_string(expr.strip())
    for func_call in funcs:
        # no "." then can't be a module
        if "." not in func_call:
            continue

        subpackage, func = func_call.rsplit(".", 1)
        package = subpackage.split(".")[0]

        # import the subpackage
        for name, short_name in aliases.items():
            subpackage = subpackage.replace(short_name, name)

        # handle the aliases
        package_import = package
        for name, short_name in aliases.items():
            if package == short_name:
                package_import = name

        # build globals
        try:
            importlib.import_module(subpackage, package=__package__)

            globs = globs | {
                package: importlib.import_module(package_import),
            }
        except Exception:
            msg = f"Function {package_import} cannot be imported"
            log.debug(msg)
            continue

    return expr, globs


def get_channels_from_groups(names: list | str | None, groupings: dict | None = None) -> list:
    """Get a list of channels from a list of groups.

    Parameters
    ----------
    names
        list of channel names
    groupings
        dictionary of the groupings of channels

    Returns
    -------
    list of channels
    """
    if names is None:
        channels_e = []
    elif isinstance(names, str):
        channels_e = groupings[names]
    elif isinstance(names, list):
        channels_e = list(itertools.chain.from_iterable([groupings[e] for e in names]))
    else:
        msg = f"names {names} must be list or str or `None`"
        raise ValueError(msg)

    return channels_e


def merge_dicts(dict_list: list) -> dict:
    """Merge a list of dictionaries, concatenating the items where they exist.

    Parameters
    ----------
    dict_list
        list of dictionaries to merge

    Returns
    -------
    a new dictionary after merging.

    Examples
    --------
    >>> merge_dicts([{"a":[1,2,3],"b":[2]},{"a":[4,5,6],"c":[2]}])
    {"a":[1,2,3,4,5,6],"b":[2],"c":[2]}
    """
    merged = {}

    for tmp_dict in dict_list:
        for key, item in tmp_dict.items():
            if key in merged:
                merged[key].extend(item)
            else:
                merged[key] = item

    return merged


@contextmanager
def filter_logging(level):
    logger = logging.getLogger("root")
    old_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(old_level)


def _check_input_file(parser, file: str | Iterable[str], descr: str = "input") -> None:
    file = (file,) if isinstance(file, str) else file
    not_existing = [f for f in file if not Path(f).exists()]
    if not_existing != []:
        parser.error(f"{descr} file(s) {''.join(not_existing)} missing")


def _check_output_file(parser, file: str | Iterable[str] | None, optional: bool = False) -> None:
    if file is None and optional:
        return

    file = (file,) if isinstance(file, str) else file
    for f in file:
        if Path(f).exists():
            parser.error(f"output file {f} already exists")



from __future__ import annotations

from pathlib import Path

from lgdo import lh5

from reboost.cli import cli


def test_cli(tmptestdir):
    test_file_dir = Path(__file__).parent / "hit"

    # test cli for build_glm
    cli(
        [
            "build-glm",
            "--id-name",
            "evtid",
            "-w",
            "--glm-file",
            f"{tmptestdir}/glm.lh5",
            "--stp-file",
            f"{test_file_dir}/test_files/beta_small.lh5",
        ]
    )

    glm = lh5.read("glm/det001", f"{tmptestdir}/glm.lh5").view_as("ak")
    assert glm.fields == ["evtid", "n_rows", "start_row"]

    # test cli for build_hit
    cli(
        [
            "build-hit",
            "--config",
            f"{test_file_dir}/configs/hit_config.yaml",
            "-w",
            "--glm-file",
            f"{tmptestdir}/glm.lh5",
            "--stp-file",
            f"{test_file_dir}/test_files/beta_small.lh5",
            "--hit-file",
            f"{tmptestdir}/hit.lh5",
            "--args",
            f"{test_file_dir}/configs/args.yaml",
        ]
    )

    hit1 = lh5.read("hit/det001", f"{tmptestdir}/hit.lh5").view_as("ak")
    assert set(hit1.fields) == {"evtid", "t0", "truth_energy", "active_energy", "smeared_energy"}



from __future__ import annotations

import logging

import awkward as ak
import numba
import numpy as np
import pint
import pyg4ometry
from lgdo import Array, VectorOfVectors
from numpy.typing import ArrayLike, NDArray

from .. import units
from ..units import ureg as u
from .utils import HPGeScalarRZField

log = logging.getLogger(__name__)


def r90(edep: ak.Array, xloc: ak.Array, yloc: ak.Array, zloc: ak.Array) -> Array:
    """R90 HPGe pulse shape heuristic.

    Parameters
    ----------
    edep
        array of energy.
    xloc
        array of x coordinate position.
    yloc
        array of y coordinate position.
    zloc
        array of z coordinate position.
    """
    tot_energy = ak.sum(edep, axis=-1, keepdims=True)

    def eweight_mean(field, energy):
        return ak.sum(energy * field, axis=-1, keepdims=True) / tot_energy

    # Compute distance of each edep to the weighted mean
    dist = np.sqrt(
        (xloc - eweight_mean(edep, xloc)) ** 2
        + (yloc - eweight_mean(edep, yloc)) ** 2
        + (zloc - eweight_mean(edep, zloc)) ** 2
    )

    # Sort distances and corresponding edep within each event
    sorted_indices = ak.argsort(dist, axis=-1)
    sorted_dist = dist[sorted_indices]
    sorted_edep = edep[sorted_indices]

    def _ak_cumsum(layout, **_kwargs):
        if layout.is_numpy:
            return ak.contents.NumpyArray(np.cumsum(layout.data))

        return None

    # Calculate the cumulative sum of energies for each event
    cumsum_edep = ak.transform(
        _ak_cumsum, sorted_edep
    )  # Implement cumulative sum over whole jagged array
    if len(edep) == 1:
        cumsum_edep_corrected = cumsum_edep
    else:
        cumsum_edep_corrected = (
            cumsum_edep[1:] - cumsum_edep[:-1, -1]
        )  # correct to get cumsum of each lower level array
        cumsum_edep_corrected = ak.concatenate(
            [
                cumsum_edep[:1],  # The first element of the original cumsum is correct
                cumsum_edep_corrected,
            ]
        )

    threshold = 0.9 * tot_energy
    r90_indices = ak.argmax(cumsum_edep_corrected >= threshold, axis=-1, keepdims=True)
    r90 = sorted_dist[r90_indices]

    return Array(ak.flatten(r90).to_numpy())


def drift_time(
    xloc: ArrayLike,
    yloc: ArrayLike,
    zloc: ArrayLike,
    dt_map: HPGeScalarRZField,
    coord_offset: pint.Quantity | pyg4ometry.gdml.Position = (0, 0, 0) * u.m,
) -> VectorOfVectors:
    """Calculates drift times for each step (cluster) in an HPGe detector.

    Parameters
    ----------
    xloc
        awkward array of x coordinate position.
    yloc
        awkward array of y coordinate position.
    zloc
        awkward array of z coordinate position.
    dt_map
        the drift time map.
    coord_offset
        this `(x, y, z)` coordinates will be subtracted to (xloc, yloc, zloc)`
        before drift time computation. The length units must be the same as
        `xloc`, `yloc` and `zloc`.
    """
    # sanitize coord_offset
    coord_offset = units.pg4_to_pint(coord_offset)

    # unit handling (for matching with drift time map units)
    xu, yu = [units.units_convfact(data, dt_map.r_units) for data in (xloc, yloc)]
    zu = units.units_convfact(zloc, dt_map.z_units)

    # unwrap LGDOs
    xloc, yloc, zloc = [units.unwrap_lgdo(data)[0] for data in (xloc, yloc, zloc)]

    # awkward transform to apply the drift time map to the step coordinates
    def _ak_dt_map(layouts, **_kwargs):
        if layouts[0].is_numpy and layouts[1].is_numpy:
            return ak.contents.NumpyArray(
                dt_map.φ(np.stack([layouts[0].data, layouts[1].data], axis=1))
            )

        return None

    # transform coordinates
    xloc = xu * xloc - coord_offset[0].to(dt_map.r_units).m
    yloc = yu * yloc - coord_offset[1].to(dt_map.r_units).m
    zloc = zu * zloc - coord_offset[2].to(dt_map.z_units).m

    # evaluate the drift time
    dt_values = ak.transform(
        _ak_dt_map,
        np.sqrt(xloc**2 + yloc**2),
        zloc,
    )

    return VectorOfVectors(
        dt_values,
        attrs={"units": units.unit_to_lh5_attr(dt_map.φ_units)},
    )


def drift_time_heuristic(
    drift_time: ArrayLike,
    edep: ArrayLike,
) -> Array:
    """HPGe drift-time-based pulse-shape heuristic.

    See :func:`_drift_time_heuristic_impl` for a description of the algorithm.

    Parameters
    ----------
    drift_time
        drift time of charges originating from steps/clusters. Can be
        calculated with :func:`drift_time`.
    edep
        energy deposited in step/cluster (same shape as `drift_time`).
    """
    # extract LGDO data and units
    drift_time, t_units = units.unwrap_lgdo(drift_time)
    edep, e_units = units.unwrap_lgdo(edep)

    # we want to attach the right units to the dt heuristic, if possible
    attrs = {}
    if t_units is not None and e_units is not None:
        attrs["units"] = units.unit_to_lh5_attr(t_units / e_units)

    return Array(_drift_time_heuristic_impl(drift_time, edep), attrs=attrs)


@numba.njit(cache=True)
def _drift_time_heuristic_impl(
    dt: ak.Array,
    edep: ak.Array,
) -> NDArray:
    r"""Low-level implementation of the HPGe drift-time-based pulse-shape heuristic.

    Accepts Awkward arrays and uses Numba to speed up the computation.

    For each hit (collection of steps), the drift times and corresponding
    energies are sorted in ascending order. The function finds the optimal
    split point :math:`m` that maximizes the *identification metric*:

    .. math::

       I = \frac{|T_1 - T_2|}{E_\text{s}(E_1, E_2)}

    where:

    .. math::

        T_1 = \frac{\sum_{i < m} t_i E_i}{\sum_{i < m} E_i}
        \quad \text{and} \quad
        T_2 = \frac{\sum_{i \geq m} t_i E_i}{\sum_{i \geq m} E_i}

    are the energy-weighted mean drift times of the two groups.

    .. math::

        E_\text{scale}(E_1, E_2) = \frac{1}{\sqrt{E_1 E_2}}

    is the scaling factor.

    The function iterates over all possible values of :math:`m` and selects the
    maximum `I` as the drift time heuristic value.
    """
    dt_heu = np.zeros(len(dt))

    # loop over hits
    for i in range(len(dt)):
        t = np.asarray(dt[i])
        e = np.asarray(edep[i])

        valid_idx = np.where(e > 0)[0]
        if len(valid_idx) < 2:
            continue

        t = t[valid_idx]
        e = e[valid_idx]

        sort_idx = np.argsort(t)
        t = t[sort_idx]
        e = e[sort_idx]

        max_id_metric = 0
        for j in range(1, len(t)):
            e1 = np.sum(e[:j])
            e2 = np.sum(e[j:])

            t1 = np.sum(t[:j] * e[:j]) / e1
            t2 = np.sum(t[j:] * e[j:]) / e2

            id_metric = abs(t1 - t2) * np.sqrt(e1 * e2)

            max_id_metric = max(max_id_metric, id_metric)

        dt_heu[i] = max_id_metric

    return dt_heu



from __future__ import annotations

import logging
import time
import typing

import awkward as ak
from lgdo.lh5 import LH5Store
from lgdo.types import LGDO, Table

from reboost import build_glm

log = logging.getLogger(__name__)


class GLMIterator:
    """A class to iterate over the rows of an event lookup map."""

    def __init__(
        self,
        glm_file: str | None,
        stp_file: str,
        lh5_group: str,
        start_row: int,
        n_rows: int | None,
        *,
        stp_field: str = "stp",
        read_vertices: bool = False,
        buffer: int = 10000,
        time_dict: dict | None = None,
    ):
        """Constructor for the glmIterator.

        Parameters
        ----------
        glm_file
            the file containing the event lookup map, if `None` the glm will
            be created in memory.
        stp_file
            the file containing the steps to read.
        lh5_group
            the name of the lh5 group to read.
        start_row
            the first row to read.
        n_rows
            the number of rows to read, if `None` read them all.
        stp_field
            name of the group.
        read_vertices
            whether to read also the vertices table.
        buffer
            the number of rows to read at once.
        time_dict
            time profiling data structure.
        """
        # initialise
        self.glm_file = glm_file
        self.stp_file = stp_file
        self.lh5_group = lh5_group
        self.start_row = start_row
        self.start_row_tmp = start_row
        self.n_rows = n_rows
        self.buffer = buffer
        self.current_i_entry = 0
        self.read_vertices = read_vertices
        self.stp_field = stp_field

        # would be good to replace with an iterator
        self.sto = LH5Store()
        self.n_rows_read = 0
        self.time_dict = time_dict
        self.glm = None

        # build the glm in memory
        if self.glm_file is None:
            self.glm = build_glm.build_glm(stp_file, None, out_table_name="glm", id_name="evtid")

    def __iter__(self) -> typing.Iterator:
        self.current_i_entry = 0
        self.n_rows_read = 0
        self.start_row_tmp = self.start_row
        return self

    def __next__(self) -> tuple[LGDO, LGDO | None, int, int]:
        # get the number of rows to read
        if self.n_rows is not None:
            rows_left = self.n_rows - self.n_rows_read
            n_rows = self.buffer if (self.buffer > rows_left) else rows_left
        else:
            n_rows = self.buffer

        if self.time_dict is not None:
            time_start = time.time()

        # read the glm rows]
        if self.glm_file is not None:
            glm_rows, n_rows_read = self.sto.read(
                f"/glm/{self.lh5_group}", self.glm_file, start_row=self.start_row_tmp, n_rows=n_rows
            )
        else:
            # get the maximum row to read
            max_row = self.start_row_tmp + n_rows
            max_row = min(len(self.glm[self.lh5_group]), max_row)

            if max_row != self.start_row_tmp:
                glm_rows = Table(self.glm[self.lh5_group][self.start_row_tmp : max_row])

            n_rows_read = max_row - self.start_row_tmp

        if self.time_dict is not None:
            self.time_dict.update_field("read/glm", time_start)

        self.n_rows_read += n_rows_read
        self.start_row_tmp += n_rows_read

        if n_rows_read == 0:
            raise StopIteration

        # view our glm as an awkward array
        glm_ak = glm_rows.view_as("ak")

        # remove empty rows
        glm_ak = glm_ak[glm_ak.n_rows > 0]

        if len(glm_ak) > 0:
            # extract range of stp rows to read
            start = glm_ak.start_row[0]
            n = ak.sum(glm_ak.n_rows)

            if self.time_dict is not None:
                time_start = time.time()

            stp_rows, n_steps = self.sto.read(
                f"/{self.stp_field}/{self.lh5_group}",
                self.stp_file,
                start_row=int(start),
                n_rows=int(n),
            )

            # save time
            if self.time_dict is not None:
                self.time_dict.update_field("read/stp", time_start)

            self.current_i_entry += 1

            if self.read_vertices:
                vert_rows, _ = self.sto.read(
                    "/vtx",
                    self.stp_file,
                    start_row=self.start_row,
                    n_rows=n_rows,
                )
            else:
                vert_rows = None
            # vertex table should have same structure as glm

            return (stp_rows, vert_rows, self.current_i_entry - 1, n_steps)
        return (None, None, self.current_i_entry, 0)



from __future__ import annotations

import awkward as ak
import numpy as np
import pyg4ometry
import pytest
from legendhpges import make_hpge
from legendtestdata import LegendTestData
from lgdo import types
from lgdo.types import VectorOfVectors

from reboost.hpge.surface import distance_to_surface


@pytest.fixture(scope="session")
def test_data_configs():
    ldata = LegendTestData()
    ldata.checkout("5f9b368")
    return ldata.get_path("legend/metadata/hardware/detectors/germanium/diodes")


def test_distance_to_surface(test_data_configs):
    gedet = make_hpge(test_data_configs + "/V99000A.json", registry=pyg4ometry.geant4.Registry())
    dist = [100, 0, 0]

    pos = ak.Array(
        {
            "xloc": [[0, 100, 200], [100], [700, 500, 200]],
            "yloc": [[100, 0, 0], [200], [100, 300, 200]],
            "zloc": [[700, 10, 20], [100], [300, 100, 0]],
            "distance_to_surface": [[1, 1, 1], [10], [1, 1, 1]],
        }
    )

    # check just the shape
    assert ak.all(
        ak.num(
            distance_to_surface(
                pos.xloc, pos.yloc, pos.zloc, gedet, det_pos=dist, surface_type=None
            ),
            axis=1,
        )
        == [3, 1, 3]
    )

    # check it can be written
    dist_full = distance_to_surface(
        pos.xloc, pos.yloc, pos.zloc, gedet, det_pos=dist, surface_type=None
    )
    assert isinstance(dist_full, types.LGDO)

    # check skipping the calculation for points > 5 mm
    dist = distance_to_surface(
        VectorOfVectors(pos.xloc),
        VectorOfVectors(pos.yloc),
        VectorOfVectors(pos.zloc),
        gedet,
        det_pos=dist,
        surface_type=None,
        distances_precompute=VectorOfVectors(pos.distance_to_surface),
        precompute_cutoff=5,
    )

    assert isinstance(dist, types.LGDO)

    assert ak.all(dist[0] == dist_full[0])
    assert ak.all(dist[2] == dist_full[2])

    assert np.isnan(dist[1][0])



from __future__ import annotations

import logging

import awkward as ak
import numpy as np
from lgdo import Array
from numpy.typing import ArrayLike

log = logging.getLogger(__name__)


def gaussian_sample(mu: ArrayLike, sigma: ArrayLike | float, *, seed: int = 999) -> Array:
    r"""Generate samples from a gaussian.

    Based on:

    .. math::

        y_i \sim \mathcal{N}(\mu_i,\sigma_i)

    where $y_i$ is the output, $x_i$ the input (mu) and $\sigma$ is the standard
    deviation for each point.

    Parameters
    ----------
    mu
        the mean positions to sample from, should be a flat (ArrayLike) object.
    sigma
        the standard deviation for each input value, can also be a single float.
    seed
        the random seed.

    Returns
    -------
    sampled values.
    """
    # convert inputs

    if isinstance(mu, Array):
        mu = mu.view_as("np")
    elif isinstance(mu, ak.Array):
        mu = mu.to_numpy()
    elif not isinstance(mu, np.ndarray):
        mu = np.array(mu)

    # similar for sigma
    if isinstance(sigma, Array):
        sigma = sigma.view_as("np")
    elif isinstance(sigma, ak.Array):
        sigma = sigma.to_numpy()
    elif not isinstance(sigma, (float, int, np.ndarray)):
        sigma = np.array(sigma)

    rng = np.random.default_rng(seed=seed)  # Create a random number generator

    return Array(rng.normal(loc=mu, scale=sigma))



from __future__ import annotations

import importlib.util
import types

from numba import njit


def numba_pdgid_funcs():
    """Load a numby-optimized copy of the scikit-hep/particle package."""
    spec_pdg = importlib.util.find_spec("particle.pdgid.functions")
    pdg_func = importlib.util.module_from_spec(spec_pdg)
    spec_pdg.loader.exec_module(pdg_func)

    def _digit2(pdgid, loc: int) -> int:
        e = 10 ** (loc - 1)
        return (pdgid // e % 10) if pdgid >= e else 0

    pdg_func._digit = _digit2

    for fname, f in pdg_func.__dict__.items():
        if not callable(f) or not isinstance(f, types.FunctionType):
            continue
        setattr(pdg_func, fname, njit(f, cache=True))

    return pdg_func



from __future__ import annotations

import awkward as ak
import numpy as np
import pytest
from lgdo import Table, lh5

from reboost.build_glm import build_glm, get_glm_rows, get_stp_evtids
from reboost.build_hit import GLMIterator


# test the basic (awkward operations) to get the glm rows
def test_get_glm_rows():
    # some basic cases
    stp_evtids = [0, 0, 0, 1, 1, 1, 6, 6, 7]
    vert = [0, 1, 6, 7]
    glm = get_glm_rows(stp_evtids, vert, start_row=0)

    assert ak.all(glm.evtid == [0, 1, 6, 7])
    assert ak.all(glm.n_rows == [3, 3, 2, 1])
    assert ak.all(glm.start_row == [0, 3, 6, 8])

    # test with a different start row

    glm = get_glm_rows(stp_evtids, vert, start_row=999)
    assert ak.all(glm.evtid == [0, 1, 6, 7])
    assert ak.all(glm.n_rows == [3, 3, 2, 1])
    assert ak.all(glm.start_row == [999, 1002, 1005, 1007])

    # test discard steps

    vert = [1, 6]
    glm = get_glm_rows(stp_evtids, vert, start_row=0)
    assert ak.all(glm.evtid == [1, 6])
    assert ak.all(glm.n_rows == [3, 2])
    assert ak.all(glm.start_row == [3, 6])

    # test gracefully fails

    # steps not in the vert table will cause it to fail
    with pytest.raises(ValueError):
        get_glm_rows(stp_evtids=[1, 3], vert=[1, 2, 4])

    # steps must be sorted
    with pytest.raises(ValueError):
        get_glm_rows(stp_evtids=[1, 3, 2], vert=[1, 2])

    # vertex evtids must be sorted
    with pytest.raises(ValueError):
        get_glm_rows(stp_evtids=[1, 3, 2], vert=[1, 2, 0])


# create some example inputs
def test_data_files(tmptestdir):
    rng = np.random.default_rng()

    # simple every evtid in vertices
    vertex_evtid = ak.Array({"evtid": np.arange(10000)})
    lh5.write(Table(vertex_evtid), "vtx", tmptestdir / "simple_test.lh5", wo_mode="of")

    # make some simple stp file
    steps_1 = ak.Array({"evtid": np.sort(rng.integers(0, 10000, size=21082))})
    steps_2 = ak.Array({"evtid": np.sort(rng.integers(0, 1000, size=1069))})

    lh5.write(Table(steps_1), "stp/det1", tmptestdir / "simple_test.lh5", wo_mode="append")
    lh5.write(Table(steps_2), "stp/det2", tmptestdir / "simple_test.lh5", wo_mode="append")

    # file with some gaps (multithreaded mode)

    vertex_evtid = ak.Array({"evtid": np.sort(np.unique(rng.integers(0, 200000, size=10000)))})
    lh5.write(Table(vertex_evtid), "vtx", tmptestdir / "gaps_test.lh5", wo_mode="of")

    # make some simple stp file
    steps_1 = ak.Array({"evtid": np.sort(rng.choice(vertex_evtid.evtid, size=21082))})
    steps_2 = ak.Array({"evtid": np.sort(rng.choice(vertex_evtid.evtid, size=1069))})

    lh5.write(Table(steps_1), "stp/det1", tmptestdir / "gaps_test.lh5", wo_mode="append")
    lh5.write(Table(steps_2), "stp/det2", tmptestdir / "gaps_test.lh5", wo_mode="append")


def test_read_stp_rows(tmptestdir):
    # check reading from the start everything
    start_row, chunk_start, evtids = get_stp_evtids(
        "stp/det1",
        str(tmptestdir / "simple_test.lh5"),
        "evtid",
        start_row=0,
        last_vertex_evtid=10000,
        stp_buffer=1000,
    )
    # read the evtid directly to compare
    evtids_read = lh5.read_as("stp/det1/evtid", str(tmptestdir / "simple_test.lh5"), "np")
    assert chunk_start == 0
    assert np.all(evtids == evtids_read)

    # check the breaking
    # the number of evtids less than 1200 is index
    index = sum(evtids_read < 1200)
    start_row, chunk_start, evtids = get_stp_evtids(
        "stp/det1",
        str(tmptestdir / "simple_test.lh5"),
        "evtid",
        start_row=0,
        last_vertex_evtid=1200,
        stp_buffer=1000,
    )
    # check we read far enough
    assert len(evtids) > index
    assert chunk_start == 0
    assert start_row == np.floor(index / 1000) * 1000

    # check updated start row

    start_row, chunk_start, evtids = get_stp_evtids(
        "stp/det1",
        str(tmptestdir / "simple_test.lh5"),
        "evtid",
        start_row=300,
        last_vertex_evtid=10000,
        stp_buffer=1000,
    )
    # first chunk should be 300
    assert chunk_start == 300
    assert len(evtids) == len(evtids_read) - 300

    # check reading of last chunk

    start_row, chunk_start, evtids = get_stp_evtids(
        "stp/det1",
        str(tmptestdir / "simple_test.lh5"),
        "evtid",
        start_row=21050,
        last_vertex_evtid=10000,
        stp_buffer=1000,
    )
    assert len(evtids) == len(evtids_read) - 21050
    assert start_row == 21050


def test_build_glm(tmptestdir):
    # produce directly glm without iteration
    # try with different buffers

    for buffer in [71, 100, 1000, 2000, 40000]:
        # two files (no gaps and gaps)
        for test in ["simple", "gaps"]:
            evtids = lh5.read_as("vtx/evtid", str(tmptestdir / f"{test}_test.lh5"), "np")

            evtids1_read = lh5.read_as("stp/det1/evtid", str(tmptestdir / f"{test}_test.lh5"), "np")
            evtids2_read = lh5.read_as("stp/det2/evtid", str(tmptestdir / f"{test}_test.lh5"), "np")
            # check both returning and saving
            for glm_file in [str(tmptestdir / f"{test}_glm.lh5"), None]:
                glm = build_glm(
                    str(tmptestdir / f"{test}_test.lh5"),
                    glm_file,
                    id_name="evtid",
                    evtid_buffer=1000,
                    stp_buffer=buffer,
                )

                if glm_file is not None:
                    glm1 = lh5.read("glm/det1", glm_file).view_as("ak")
                    glm2 = lh5.read("glm/det2", glm_file).view_as("ak")
                    glm = ak.Array({"det1": glm1, "det2": glm2})
                # glm should have the right evtid

                assert ak.all(glm.det1.evtid.to_numpy() == evtids)
                assert ak.all(glm.det2.evtid.to_numpy() == evtids)

                # total number of rows should be correct
                assert np.sum(glm.det1.n_rows) == len(evtids1_read)
                assert np.sum(glm.det2.n_rows) == len(evtids2_read)


def test_glm_iterator(tmptestdir):
    # make an glm

    # two files (no gaps and gaps)
    for test in ["simple", "gaps"]:
        stp_file = str(tmptestdir / f"{test}_test.lh5")
        glm_file = str(tmptestdir / f"{test}_glm.lh5")

        build_glm(
            stp_file,
            glm_file,
            id_name="evtid",
            evtid_buffer=1000,
            stp_buffer=100000,
        )

        # iterate over the glm and the test file
        # consider both a chunked and a full read
        # also try with the glm file being None
        for gt in [glm_file, None]:
            for n_rows in [5000, None]:
                for det in ["det1", "det2"]:
                    evtids = None
                    glm_it = GLMIterator(
                        gt,
                        stp_file,
                        lh5_group=det,
                        start_row=0,
                        stp_field="stp",
                        n_rows=n_rows,
                        read_vertices=True,
                        buffer=100,
                    )
                    # get the overall evtids
                    for stps, _, _, _ in glm_it:
                        if stps is None:
                            continue

                        evtids = (
                            stps.view_as("ak").evtid
                            if evtids is None
                            else ak.concatenate((evtids, stps.view_as("ak").evtid))
                        )

                    evtids_read = lh5.read_as(
                        f"stp/{det}/evtid", str(tmptestdir / f"{test}_test.lh5"), "np"
                    )

                    assert ak.all(evtids == evtids_read)



from __future__ import annotations

import logging
import time
from typing import Any

import awkward as ak
from dbetto import AttrsDict
from lgdo.types import LGDO, Table

from reboost.profile import ProfileDict

from . import utils

log = logging.getLogger(__name__)


def evaluate_output_column(
    hit_table: Table,
    expression: str,
    local_dict: dict,
    *,
    table_name: str = "HITS",
    time_dict: ProfileDict | None = None,
    name: str = " ",
) -> LGDO:
    """Evaluate an expression returning an LGDO.

    Uses :func:`lgdo.Table.eval()` to compute a new column for the
    hit table. The expression can depend on any field in the Table
    (prefixed with table_name.) or objects contained in the local dict.
    In addition, the expression can use packages which are then imported.

    Parameters
    ----------
    hit_table
        the table containing the hit fields.
    expression
        the expression to evaluate.
    local_dict
        local dictionary to pass to :func:`lgdo.Table.eval()`.
    table_name
        keyword used to refer to the fields in the table.
    time_dict
        time profiling data structure.
    name
        name to use in `time_dict`.

    Returns
    -------
    an LGDO with the new field.
    """
    if time_dict is not None:
        time_start = time.time()

    if local_dict is None:
        local_dict = {}

    expr = expression.replace(f"{table_name}.", "")

    # get func call and modules to import

    func_call, globals_dict = utils.get_function_string(expr)

    msg = f"evaluating table with command {expr} and local_dict {local_dict.keys()}"
    log.debug(msg)

    # remove np and ak
    globals_dict.pop("np", None)
    globals_dict.pop("ak", None)

    if globals_dict == {}:
        globals_dict = None

    if globals_dict is not None and "pyg4ometry" in globals_dict:
        with utils.filter_logging(logging.CRITICAL):
            res = hit_table.eval(func_call, local_dict, modules=globals_dict)
    else:
        res = hit_table.eval(func_call, local_dict, modules=globals_dict)

    # how long did it take
    if time_dict is not None:
        time_dict.update_field(name=f"expressions/{name}", time_start=time_start)

    return res


def evaluate_object(
    expression: str,
    local_dict: dict,
) -> Any:
    """Evaluate an expression returning any object.

    The expression should be a function call. It can depend on any objects contained in the local dict.
    In addition, the expression can use packages which are then imported.

    Parameters
    ----------
    expression
        the expression to evaluate.
    local_dict
        local dictionary to pass to `eval()`.

    Returns
    -------
    the evaluated object.
    """
    msg = f"Evaluating object with expression {expression} and {local_dict}"
    log.debug(msg)

    func_call, globals_dict = utils.get_function_string(expression)

    if "pyg4ometry" in globals_dict:
        with utils.filter_logging(logging.CRITICAL):
            return eval(func_call, local_dict, globals_dict)
    else:
        return eval(func_call, local_dict, globals_dict)


def get_global_objects(
    expressions: dict[str, str], *, local_dict: dict, time_dict: dict | None = None
) -> dict:
    """Extract global objects used in the processing.

    Parameters
    ----------
    expressions
        a dictionary containing the expressions to evaluate for each object.
    local_dict
        other objects used in the evaluation of the expressions, passed to
        `eval()` as the locals keyword.
    time_dict
        time profiling data structure.

    Returns
    -------
    dictionary of objects with the same keys as the expressions.
    """
    if time_dict is not None:
        time_start = time.time()

    msg = f"Getting global objects with {expressions.keys()} and {local_dict}"
    log.info(msg)

    res = AttrsDict(
        {
            obj_name: evaluate_object(expression, local_dict=local_dict)
            for obj_name, expression in expressions.items()
        }
    )
    if time_dict is not None:
        time_dict.update_field(name="global_objects", time_start=time_start)
    return res


def get_detectors_mapping(
    output_detector_expression: str | list,
    objects: AttrsDict | None = None,
    input_detector_name: str | None = None,
) -> dict:
    """Extract the output detectors and the list of input to outputs by parsing the expressions.

    The output_detector_expression can be a name or a string evaluating to a list of names.
    This expression can depend on any objects in the objects dictionary, referred to by the keyword
    "OBJECTS".

    The function produces a dictionary mapping input detectors to output detectors with the following
    format:

    .. code-block:: python

        {
            "input1": ["output1", "output2"],
            "input2": ["ouput3", ...],
        }

    If only output_detector_expression is supplied the mapping is one-to-one (i.e. every
    input detector maps to the same output detector). If instead a name for the input_detector_name
    is also supplied this will be the only key with all output detectors being mapped to this.

    Parameters
    ----------
    output_detector_expression
        An output detector name or a string evaluating to a list of output tables.
    objects
        dictionary of objects that can be referenced in the expression.
    input_detector_name
        Optional input detector name for all the outputs.


    Returns
    -------
    a dictionary with the input detectors as key and a list of output detectors for each.

    Examples
    --------
    For a direct one-to-one mapping:

    >>> get_detectors_mapping("[str(i) for i in range(2)]")
    {'0':['0'],'1':['1'],'2':['2']}

    With an input detector name:

    >>> get_detectors_mapping("[str(i) for i in range(2)])",input_detector_name = "dets")
    {'dets':['0','1','2']}

    With  objects:

    >>> objs = AttrsDict({"format": "ch"})
    >>> get_detectors_mapping("[f'{OBJECTS.format}{i}' for i in range(2)])",
                                input_detector_name = "dets",objects=objs)
    {'dets': ['ch0', 'ch1', 'ch2']}
    """
    out_names = []
    if isinstance(output_detector_expression, str):
        out_list = [output_detector_expression]
    else:
        out_list = list(output_detector_expression)

    for expression_tmp in out_list:
        func, globs = utils.get_function_string(expression_tmp)

        # if no package was imported its just a name
        try:
            objs = evaluate_object(expression_tmp, local_dict={"OBJECTS": objects})
            out_names.extend(objs)
        except Exception:
            out_names.append(expression_tmp)

    # simple one to one mapping
    if input_detector_name is None:
        return {name: [name] for name in out_names}
    return {input_detector_name: out_names}


def get_detector_objects(
    output_detectors: list,
    expressions: dict,
    args: AttrsDict,
    global_objects: AttrsDict,
    time_dict: ProfileDict | None = None,
) -> AttrsDict:
    """Get the detector objects for each detector.

    This computes a set of objects per output detector. These should be the
    expressions (defined in the `expressions` input). They can depend
    on the keywords:

    - `ARGS` : in which case values of from the args parameter AttrsDict can be references,
    - `DETECTOR`: referring to the detector name (key of the detector mapping)
    - `OBJECTS` : The global objects.

    For example expressions like:

    .. code-block:: python

        compute_object(arg=ARGS.first_arg, detector=DETECTOR, obj=OBJECTS.meta)

    are supported.

    Parameters
    ----------
    output_detectors
        list of output detectors,
    expressions
        dictionary of expressions to evaluate.
    args
        any arguments the expression can depend on, is passed as `locals` to `eval()`.
    global_objects
        a dictionary of objects the expression can depend on.
    time_dict
        time profiling data structure.

    Returns
    -------
    An AttrsDict of the objects for each detector.
    """
    if time_dict is not None:
        time_start = time.time()

    det_objects_dict = {}
    for output_detector in output_detectors:
        obj_dict = {}
        for obj_name, obj_expression in expressions.items():
            obj_dict[obj_name] = evaluate_object(
                obj_expression,
                local_dict={
                    "ARGS": args,
                    "DETECTOR": output_detector,
                    "OBJECTS": global_objects,
                    "DETECTOR_OBJECTS": AttrsDict(obj_dict),
                },
            )

        det_objects_dict[output_detector] = AttrsDict(obj_dict)
    res = AttrsDict(det_objects_dict)

    if time_dict is not None:
        time_dict.update_field(name="detector_objects", time_start=time_start)

    return res


def evaluate_hit_table_layout(
    steps: ak.Array | Table, expression: str, time_dict: dict | None = None
) -> Table:
    """Evaluate the hit_table_layout expression, producing the hit table.

    This expression should be a function call which performs a restructuring of the steps,
    i.e. it sets the number of rows. The steps array should be referred to
    by "STEPS" in the expression.

    Parameters
    ----------
    steps
        awkward array or Table of the steps.
    expression
        the expression to evaluate to produce the hit table.
    time_dict
        time profiling data structure.

    Returns
    -------
    :class:`lgdo.Table` of the hits.
    """
    if time_dict is not None:
        time_start = time.time()

    group_func, globs = utils.get_function_string(
        expression,
    )
    locs = {"STEPS": steps}

    msg = f"running step grouping with {group_func} and globals {globs.keys()} and locals {locs.keys()}"
    log.debug(msg)

    res = eval(group_func, globs, locs)

    if time_dict is not None:
        time_dict.update_field(name="hit_layout", time_start=time_start)

    return res


def remove_columns(tab: Table, outputs: list) -> Table:
    """Remove columns from the table not found in the outputs.

    Parameters
    ----------
    tab
        the table to remove columns from.
    outputs
        a list of output fields.

    Returns
    -------
    the table with columns removed.
    """
    existing_cols = list(tab.keys())
    for col in existing_cols:
        if col not in outputs:
            tab.remove_column(col, delete=True)

    return tab


def merge(hit_table: Table, output_table: ak.Array | None):
    """Merge the table with the array."""
    return (
        hit_table.view_as("ak")
        if output_table is None
        else ak.concatenate((output_table, hit_table.view_as("ak")))
    )



from __future__ import annotations

import logging

import awkward as ak
import numpy as np
from lgdo import Array, VectorOfVectors
from lgdo.types import LGDO

log = logging.getLogger(__name__)


def piecewise_linear_activeness(
    distances: VectorOfVectors | ak.Array, fccd: float, dlf: float
) -> VectorOfVectors | Array:
    r"""Piecewise linear HPGe activeness model.

    Based on:

    .. math::

        f(d) =
        \begin{cases}
        0 & \text{if } d < f*l, \\
        \frac{x-f*l}{f - f*l} & \text{if } t \leq d < f, \\
        1 & \text{otherwise.}
        \end{cases}

    Where:

    - `d`: Distance to surface,
    - `l`: Dead layer fraction, the fraction of the FCCD which is fully inactive
    - `f`: Full charge collection depth (FCCD).

    In addition, any distance of `np.nan` (for example if the calculation
    was not performed for some steps) is assigned an activeness of one.

    Parameters
    ----------
    distances
        the distance from each step to the detector surface. Can be either a
        `numpy` or `awkward` array, or a LGDO `VectorOfVectors` or `Array`. The computation
        is performed for each element and the shape preserved in the output.

    fccd
        the value of the FCCD
    dlf
        the fraction of the FCCD which is fully inactive.

    Returns
    -------
    a :class:`VectorOfVectors` or :class:`Array` of the activeness
    """
    # convert to ak
    if isinstance(distances, LGDO):
        distances_ak = distances.view_as("ak")
    elif not isinstance(distances, ak.Array):
        distances_ak = ak.Array(distances)
    else:
        distances_ak = distances

    dl = fccd * dlf
    distances_flat = (
        ak.flatten(distances_ak).to_numpy() if distances_ak.ndim > 1 else distances_ak.to_numpy()
    )

    # compute the linear piecewise
    results = np.full_like(distances_flat, np.nan, dtype=np.float64)
    lengths = ak.num(distances_ak) if distances_ak.ndim > 1 else len(distances_ak)

    mask1 = (distances_flat > fccd) | np.isnan(distances_flat)
    mask2 = (distances_flat <= dl) & (~mask1)
    mask3 = ~(mask1 | mask2)

    # assign the values
    results[mask1] = 1
    results[mask2] = 0
    results[mask3] = (distances_flat[mask3] - dl) / (fccd - dl)

    # reshape
    results = ak.unflatten(ak.Array(results), lengths) if distances_ak.ndim > 1 else results

    return VectorOfVectors(results) if results.ndim > 1 else Array(results)


def vectorised_active_energy(
    distances: VectorOfVectors | ak.Array,
    edep: VectorOfVectors | ak.Array,
    fccd: float | list,
    dlf: float | list,
) -> VectorOfVectors | Array:
    r"""Energy after piecewise linear HPGe activeness model vectorised over FCCD or dead layer fraction.

    Based on the same linear activeness function as :func:`piecewise_linear_activeness`. However,
    this function vectorises the calculation to provide a range of output energies varying the fccd or
    dead layer fraction. Either fccd or dlf can be a list. This adds an extra dimension to the
    output, with the same length as the input fccd or dlf list.

    .. warning:
        It is not currently implemented to vary both dlf and fccd.

    Parameters
    ----------
    distances
        the distance from each step to the detector surface. Can be either a
        `awkward` array, or a LGDO `VectorOfVectors` . The computation
        is performed for each element and the first dimension is preserved, a
        new dimension is added vectorising over the FCCD or DLF.
    edep
        the energy for each step.
    fccd
        the value of the FCCD, can be a list.
    dlf
        the fraction of the FCCD which is fully inactive, can be a list.

    Returns
    -------
    a :class:`VectorOfVectors` or :class:`Array` of the activeness
    """
    # add checks on fccd, dlf
    fccd = np.array(fccd)
    dlf = np.array(dlf)

    if (fccd.ndim + dlf.ndim) > 1:
        msg = "Currently only one of FCCD and dlf can be varied"
        raise NotImplementedError(msg)

    # convert fccd and or dlf to the right shape
    if fccd.ndim == 0:
        if dlf.ndim == 0:
            dlf = dlf[np.newaxis]
        fccd = np.full_like(dlf, fccd)

    dl = fccd * dlf

    def _convert(field):
        # convert to ak
        if isinstance(field, VectorOfVectors):
            field_ak = field.view_as("ak")
        elif not isinstance(field, ak.Array):
            field_ak = ak.Array(field)
        else:
            msg = f"{field} must be an awkward array or VectorOfVectors"
            raise TypeError(msg)

        return field_ak, ak.flatten(field_ak).to_numpy()[:, np.newaxis]

    distances_ak, distances_flat = _convert(distances)
    _, edep_flat = _convert(edep)
    runs = ak.num(distances_ak, axis=-1)

    # vectorise fccd or tl

    fccd_list = np.tile(fccd, (len(distances_flat), 1))
    dl_list = np.tile(dl, (len(distances_flat), 1))
    distances_shaped = np.tile(distances_flat, (1, len(dl)))

    # compute the linear piecewise
    results = np.full_like(fccd_list, np.nan, dtype=np.float64)

    # Masks
    mask1 = (distances_shaped > fccd_list) | np.isnan(distances_shaped)
    mask2 = ((distances_shaped <= dl_list) | (fccd_list == dl_list)) & ~mask1
    mask3 = ~(mask1 | mask2)  # Safe, avoids recomputing anything expensive

    # Assign values
    results[mask1] = 1.0
    results[mask2] = 0.0
    results[mask3] = (distances_shaped[mask3] - dl_list[mask3]) / (
        fccd_list[mask3] - dl_list[mask3]
    )

    energy = ak.sum(ak.unflatten(results * edep_flat, runs), axis=-2)

    return VectorOfVectors(energy) if energy.ndim > 1 else Array(energy.to_numpy())



from __future__ import annotations

import legendhpges as hpges
import numpy as np
import pyg4ometry as pg4
from legendtestdata import LegendTestData

ldata = LegendTestData()

reg = pg4.geant4.Registry()

natge = hpges.materials.make_natural_germanium(reg)
hpge_l = hpges.make_hpge(
    ldata["legend/metadata/hardware/detectors/germanium/diodes/V99000A.yaml"],
    name="V99000A",
    registry=reg,
    material=natge,
)

# create a world volume
world_s = pg4.geant4.solid.Orb("World_s", 20, registry=reg, lunit="cm")
world_l = pg4.geant4.LogicalVolume(world_s, "G4_Galactic", "World", registry=reg)
reg.setWorld(world_l)

pg4.geant4.PhysicalVolume([0, 0, 0], [0, 0, 0, "m"], hpge_l, "V99000A", world_l, registry=reg)

# finally create a small radioactive source
source_s = pg4.geant4.solid.Tubs("Source_s", 0, 1, 1, 0, 2 * np.pi, registry=reg)
source_l = pg4.geant4.LogicalVolume(source_s, "G4_BRAIN_ICRP", "Source_L", registry=reg)
pg4.geant4.PhysicalVolume([0, 0, 0], [0, 0, -1, "cm"], source_l, "Source", world_l, registry=reg)

# viewer = pg4.visualisation.VtkViewerColoured()
# viewer.addLogicalVolume(reg.getWorldVolume())
# viewer.view()

w = pg4.gdml.Writer()
w.addDetector(reg)
w.write("geometry.gdml")



from __future__ import annotations

from pathlib import Path

import awkward as ak
import dbetto
import pytest
from lgdo import Array, Struct, Table, lh5

import reboost
from reboost.build_glm import build_glm


@pytest.fixture(scope="module")
def test_gen_lh5(tmptestdir):
    # write a basic lh5 file

    stp_path = str(tmptestdir / "basic.lh5")
    glm_path = str(tmptestdir / "basic_glm.lh5")

    evtid = [0, 0, 1, 1, 1]
    edep = [100, 200, 10, 20, 300]  # keV
    time = [0, 1.5, 0.1, 2.1, 3.7]  # ns
    vertices = [0, 1]
    tab = Table({"evtid": Array(evtid), "edep": Array(edep), "time": Array(time)})
    lh5.write(tab, "stp/det1", stp_path, wo_mode="of")
    lh5.write(
        Table({"evtid": Array(vertices)}),
        "vtx",
        stp_path,
        wo_mode="append",
    )

    build_glm(stp_path, glm_path, id_name="evtid")

    return stp_path, glm_path


def test_basic(test_gen_lh5, tmptestdir):
    stp_path, glm_path = test_gen_lh5

    reboost.build_hit.build_hit(
        f"{Path(__file__).parent}/configs/basic.yaml",
        args={},
        stp_files=stp_path,
        glm_files=glm_path,
        hit_files=f"{tmptestdir}/basic_hit.lh5",
    )

    hits = lh5.read("hit/det1", f"{tmptestdir}/basic_hit.lh5").view_as("ak")

    assert ak.all(hits.energy == [300, 330])
    assert ak.all(hits.t0 == [0, 0.1])
    assert ak.all(hits.evtid[0] == [0, 0])
    assert ak.all(hits.evtid[1] == [1, 1, 1])

    # test in memory

    hits, time_dict = reboost.build_hit.build_hit(
        f"{Path(__file__).parent}/configs/basic.yaml",
        args={},
        stp_files=stp_path,
        glm_files=glm_path,
        hit_files=None,
    )

    assert ak.all(hits["det1"].energy == [300, 330])
    assert ak.all(hits["det1"].t0 == [0, 0.1])
    assert ak.all(hits["det1"].evtid[0] == [0, 0])
    assert ak.all(hits["det1"].evtid[1] == [1, 1, 1])

    assert list(time_dict.keys()) == ["global_objects", "geds"]
    assert list(time_dict["geds"].keys()) == [
        "detector_objects",
        "read",
        "conv",
        "hit_layout",
        "expressions",
    ]
    assert list(time_dict["geds"]["read"].keys()) == ["glm", "stp"]
    assert list(time_dict["geds"]["expressions"].keys()) == ["t0", "first_evtid", "energy"]


def test_full_chain(tmptestdir):
    build_glm(
        f"{Path(__file__).parent}/test_files/beta_small.lh5",
        str(tmptestdir / "beta_small_glm.lh5"),
        id_name="evtid",
    )

    args = dbetto.AttrsDict(
        {
            "gdml": f"{Path(__file__).parent}/configs/geom.gdml",
            "pars": f"{Path(__file__).parent}/configs/pars.yaml",
        }
    )

    _, time_dict = reboost.build_hit.build_hit(
        f"{Path(__file__).parent}/configs/hit_config.yaml",
        args=args,
        stp_files=f"{Path(__file__).parent}/test_files/beta_small.lh5",
        glm_files=str(tmptestdir / "beta_small_glm.lh5"),
        hit_files=f"{Path(__file__).parent}/test_files/beta_small_hit.lh5",
        overwrite=True,
    )
    hits = lh5.read("hit", f"{Path(__file__).parent}/test_files/beta_small_hit.lh5")

    assert isinstance(hits, Struct)

    assert set(hits["det001"].view_as("ak").fields) == {
        "evtid",
        "t0",
        "truth_energy",
        "active_energy",
        "smeared_energy",
    }
    assert set(hits["det002"].view_as("ak").fields) == {
        "evtid",
        "t0",
        "truth_energy",
        "active_energy",
        "smeared_energy",
    }



from __future__ import annotations

import numpy as np
import pytest
from lgdo import Array, Table, lh5

from reboost.optmap.convolve import convolve
from reboost.optmap.create import (
    check_optical_map,
    create_optical_maps,
    list_optical_maps,
    merge_optical_maps,
    rebin_optical_maps,
)
from reboost.optmap.evt import build_optmap_evt
from reboost.optmap.optmap import OpticalMap


@pytest.fixture
def tbl_hits(tmptestdir):
    evt_count = 100
    rng = np.random.default_rng(1234)
    loc = rng.uniform(size=(evt_count, 3))
    evtids = np.arange(1, evt_count + 1)

    tbl_vertices = Table(
        {
            "evtid": Array(evtids),
            "xloc": Array(loc[:, 0]),
            "yloc": Array(loc[:, 1]),
            "zloc": Array(loc[:, 2]),
            "n_part": Array(np.ones(evt_count)),
            "time": Array(np.ones(evt_count)),
        }
    )

    mask = rng.uniform(size=evt_count) < 0.2
    hit_count = np.sum(mask)

    tbl_optical = Table(
        {
            "evtid": Array(evtids[mask]),
            "det_uid": Array(np.ones(hit_count, dtype=np.int_)),
            "wavelength": Array(rng.normal(loc=400, scale=30, size=hit_count)),
            "time": Array(2 * np.ones(hit_count)),
        }
    )

    hit_file = tmptestdir / "hit.lh5"
    lh5.write(tbl_vertices, name="vtx", lh5_file=hit_file, wo_mode="overwrite_file")
    lh5.write(tbl_optical, name="stp/optical", lh5_file=hit_file, wo_mode="overwrite")
    return hit_file


def test_optmap_evt(tbl_hits, tmptestdir):
    evt_out_file = tmptestdir / "evt-out.lh5"
    build_optmap_evt(
        str(tbl_hits),
        str(evt_out_file),
        detectors=("1", "002", "003"),
        buffer_len=20,  # note: shorter window sizes (e.g. 10) do not work.
    )


@pytest.fixture
def tbl_evt_fns(tmptestdir) -> tuple[str]:
    evt_count = 100
    rng = np.random.default_rng(1234)
    loc = rng.uniform(size=(evt_count, 3))
    hits = rng.geometric(p=0.9, size=(evt_count, 3)) - 1

    tbl_evt = Table(
        {
            "xloc": Array(loc[:, 0]),
            "yloc": Array(loc[:, 1]),
            "zloc": Array(loc[:, 2]),
            "001": Array(hits[:, 0]),
            "002": Array(hits[:, 1]),
            "003": Array(hits[:, 2]),
        }
    )

    evt_file = tmptestdir / "evt.lh5"
    lh5.write(tbl_evt, name="optmap_evt", lh5_file=evt_file, wo_mode="overwrite_file")
    return (str(evt_file),)


@pytest.mark.filterwarnings("ignore::scipy.optimize._optimize.OptimizeWarning")
def test_optmap_create(tbl_evt_fns):
    settings = {
        "range_in_m": [[0, 1], [0, 1], [0, 1]],
        "bins": [10, 10, 10],
    }

    # test creation only with the summary map.
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=(),
        output_lh5_fn=None,
    )

    # test creation with all detectors.
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001", "002", "003"),
        output_lh5_fn=None,
    )

    # test creation with some detectors.
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001"),
        output_lh5_fn=None,
    )

    # test creation on multiple cores.
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001", "002", "003"),
        output_lh5_fn=None,
        n_procs=2,
    )


@pytest.mark.filterwarnings("ignore::scipy.optimize._optimize.OptimizeWarning")
def test_optmap_merge(tbl_evt_fns, tmptestdir):
    settings = {
        "range_in_m": [[0, 1], [0, 1], [0, 1]],
        "bins": [10, 10, 10],
    }

    map1_fn = str(tmptestdir / "map1.lh5")
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001", "002", "003"),
        output_lh5_fn=map1_fn,
    )
    map2_fn = str(tmptestdir / "map2.lh5")
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001", "002", "003"),
        output_lh5_fn=map2_fn,
    )

    # test in sequential mode.
    map_merged_fn = str(tmptestdir / "map-merged.lh5")
    merge_optical_maps([map1_fn, map2_fn], map_merged_fn, settings)

    # also test on multiple cores.
    map_merged_fn = str(tmptestdir / "map-merged-mp.lh5")
    merge_optical_maps([map1_fn, map2_fn], map_merged_fn, settings, n_procs=2)


@pytest.mark.filterwarnings("ignore::scipy.optimize._optimize.OptimizeWarning")
def test_optmap_rebin(tbl_evt_fns, tmptestdir):
    settings = {
        "range_in_m": [[0, 1], [0, 1], [0, 1]],
        "bins": [10, 10, 10],
    }

    map1_fn = str(tmptestdir / "map-to-rebin.lh5")
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001", "002", "003"),
        output_lh5_fn=map1_fn,
    )

    map_rebinned_fn = str(tmptestdir / "map-rebinned.lh5")
    rebin_optical_maps(map1_fn, map_rebinned_fn, factor=2)


@pytest.fixture
def tbl_edep(tmptestdir):
    evt_count = 100
    rng = np.random.default_rng(1234)
    loc = rng.uniform(size=(evt_count, 6))

    evtids = np.arange(1, evt_count + 1)

    tbl_edep = Table(
        {
            "evtid": Array(evtids),
            "particle": Array(22 * np.ones(evt_count, dtype=np.int64)),
            "edep": Array(rng.normal(loc=200, scale=2, size=evt_count)),
            "xloc_pre": Array(loc[:, 0]),
            "yloc_pre": Array(loc[:, 1]),
            "zloc_pre": Array(loc[:, 2]),
            "xloc_post": Array(loc[:, 3]),
            "yloc_post": Array(loc[:, 4]),
            "zloc_post": Array(loc[:, 5]),
            "time": Array(np.ones(evt_count)),
            "v_pre": Array(60 * np.ones(evt_count)),
            "v_post": Array(58 * np.ones(evt_count)),
        }
    )

    evt_file = tmptestdir / "edep.lh5"
    lh5.write(tbl_edep, name="/stp/x", lh5_file=evt_file, wo_mode="overwrite_file")
    return evt_file


@pytest.mark.filterwarnings("ignore::scipy.optimize._optimize.OptimizeWarning")
def test_optmap_convolve(tbl_evt_fns, tbl_edep, tmptestdir):
    settings = {
        "range_in_m": [[0, 1], [0, 1], [0, 1]],
        "bins": [2, 2, 2],
    }

    map_fn = str(tmptestdir / "map-convolve.lh5")
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001"),
        output_lh5_fn=map_fn,
    )

    out_fn = str(tmptestdir / "convolved.lh5")
    convolve(
        map_fn,
        str(tbl_edep),
        "/stp/x",
        material="lar",
        output_file=out_fn,
        buffer_len=10,
    )


@pytest.mark.filterwarnings("ignore::scipy.optimize._optimize.OptimizeWarning")
def test_optmap_save_and_load(tmptestdir, tbl_evt_fns):
    settings = {
        "range_in_m": [[0, 1], [0, 1], [0, 1]],
        "bins": [10, 10, 10],
    }

    map_fn = str(tmptestdir / "map-load.lh5")
    create_optical_maps(
        tbl_evt_fns,
        settings,
        chfilter=("001", "002", "003"),
        output_lh5_fn=map_fn,
    )

    assert list_optical_maps(map_fn) == ["_001", "_002", "_003", "all"]
    om = OpticalMap.load_from_file(map_fn, "all")
    assert isinstance(om, OpticalMap)
    om = OpticalMap.load_from_file(map_fn, "all")
    assert isinstance(om, OpticalMap)

    check_optical_map(map_fn)
